# -*- coding: utf-8 -*-
import datetime
import json

from kinder.webservice.spyne_ws.exceptions import SpyneException
from kinder.core.declaration.models import Declaration, DeclarationPrivilege
from kinder.core.children.models import ChildrenDelegate, Delegate
from kinder.core.helpers import recursive_getattr
from concentrator.change import (
    DeclarationPrivilegeChangeHelper, ChangeHelper, map_changes)
from concentrator.models import ChangeStatus
from concentrator.webservice.proxy import (
    NewApplicationProxy, UpdateApplicationProxy, UpdateProxy)
from concentrator.webservice.helpers import get_delegate
from lipetsk_specifics.rules import (WorkTypeRules,
                                     DelegateWhoHaveBenefits)
from lipetsk_specifics.models import(
    PrivilegeConfirmationAttributes, PrivilegeOwnerEnum)
from kinder.provider import features
from .django_objects_proxy import (
    LipetskDelegateProxy,
    LipetskChildrenProxy,
    DelegatePrivilegeProxy,
    PrivConfirmAttrsProxy
)
from .helpers import get_decl_priv, get_priv_data


class LipetskNewApplicationProxy(NewApplicationProxy):
    """
    Прокси объект для метода NewApplication
    """
    def _create_delegate_for_priv(self, delegate_type, who_type, children,
                                  delegate):
        """
        Проверяем по ФИ и снилсу не привязан ли уже этот представитель
        к ребенку, если нет, то создаем нового, иначе возвращаем существующего
        @param delegate: уже созданные представитель в заявке
        @param who_type: Тип описывающей владельца льготы
        @param delegate_type: Тип описывающей представителя ребенка
        """
        new_delegate = None
        map_compare = ['LastName', 'FirstName', 'Snils']
        if all(getattr(delegate_type, field_name) == getattr(who_type,
                                                             field_name)
               for field_name in map_compare):
            new_delegate = delegate
        if not new_delegate:
            new_delegate, new_delegate_extension = DelegatePrivilegeProxy(
                who_type).get()
            new_delegate.save()
            new_delegate_extension.delegate = new_delegate
            new_delegate_extension.save()
            ChildrenDelegate.objects.create(
                children=children, delegate=new_delegate)
        return new_delegate

    def _save_privileges(self, privileges, declaration, doc_number, delegate):
        """
        Согласовано с аналитикаами,
        что при подаче с портала не может быть более 1 льготы
        @param privileges:
        @param declaration:
        @param doc_number:
        @param delegate:
        @return:
        """
        if len(privileges) > 1:
            raise SpyneException(u"Не должно быть более 1 льготы с портала!!!")
        if privileges:
            owner_type = self._child_data.Benefits.WhoHaveBenefit.Type
            owner_delegate = None
            if owner_type != PrivilegeOwnerEnum.CHILDREN:
                owner_delegate = self._create_delegate_for_priv(
                    self._delegate_data,
                    self._child_data.Benefits.WhoHaveBenefit,
                    declaration.children,
                    delegate)
            for privilege in privileges:
                decl_pr = DeclarationPrivilege.objects.create(
                    declaration=declaration,
                    datetime=datetime.datetime.now(),
                    privilege=privilege,
                    doc_number=doc_number)
                data = dict(
                    privilege_owner=owner_type,
                    declaration_privilege=decl_pr,
                    portal=True,
                )
                if owner_delegate:
                    data['delegate'] = owner_delegate
                PrivilegeConfirmationAttributes.objects.create(**data)

    def _create_delegate(self):
        return LipetskDelegateProxy(self._delegate_data).get()

    def _create_children(self):
        return LipetskChildrenProxy(
            self._child_data, self._delegate_data,
            self._adaptation_program_type).get()

    def _get_work_type(self):
        return WorkTypeRules.get_system(self._child_data.ArrivalTimeType)

    def _create_declaration(self):
        return Declaration(
            date=self._submit_date,
            desired_date=self._entry_date,
            offer_other=self._edu_organizations.AllowOfferOther,
            client_id=self._external_id,
            comment=self._child_data.IssuerState,
            portal=1,
            work_type_id=self._get_work_type(),
            source=features.get_feature(
                'lipetsk_specifics.webservice.service.'
                'NewApplicationRequest.get_source')())


class LipetskDeclarationPrivilegeChangeHelper(
    DeclarationPrivilegeChangeHelper
):
    """
    Изменени порядок олучения изменений по льготам
    Фиксируем изменения только для льгот пришедших с портала
    """
    def _get_current_state(self, difference, changes):
        """
        возвращает список id льгот заявки
        @param difference:
        @param changes:
        @return:
        """
        if changes.state == ChangeStatus.WAIT:
            declarationprivilege_set = get_decl_priv(changes.declaration, True)
            curent_state = map(lambda x: x.privilege.id,
                               declarationprivilege_set)
        else:
            curent_state = map(lambda x: x[self.SYS_NAME],
                               filter(lambda x: self.SYS_NAME in x.keys(),
                                      difference))
            curent_state = curent_state[0]
        return curent_state

    def _create_new(self, new_list, changes):
        """
        создаем привязку льготы к заявлению
        @param new_list:
        @return:
        """
        if len(new_list) > 1:
            raise SpyneException(u"Не должно быть более 1 льготы с портала!!!")
        for priv_id in new_list:
            decl_pr, _create = DeclarationPrivilege.objects.get_or_create(
                declaration=changes.declaration,
                privilege_id=priv_id,
            )
            # Если данные по обладателю льготы уже есть,
            # то мы только изменим в нем ссылку на льготу,
            # иначе создадим заново.
            # Тк в липецке льгота одна, считаем что именно ее обладателя мы и
            # должны изменить.
            priv_atr_qs = PrivilegeConfirmationAttributes.objects.filter(
                declaration_privilege__declaration_id=changes.declaration)
            if priv_atr_qs:
                for attr in priv_atr_qs:
                    attr.declaration_privilege = decl_pr
                    attr.confirmed = False
                    attr.save()
            else:
                data = dict(
                    declaration_privilege=decl_pr,
                    portal=True,
                )
                owner_delegate = self._create_delegate_from_change(
                    changes) or get_delegate(changes.declaration)

                if owner_delegate:
                    data['delegate'] = owner_delegate
                    data['privilege_owner'] = (
                        DelegateWhoHaveBenefits.get_concetr(owner_delegate.type))
                PrivilegeConfirmationAttributes.objects.create(**data)

    def _create_delegate_from_change(self, changes):
        try:
            changes_dict = json.loads(changes.data)
        except ValueError:
            return None

        try:
            confirmation_attributes = changes_dict[
                "PrivilegeConfirmationAttributes"]
        except KeyError:
            return None

        data = {}
        for attribute in confirmation_attributes:
            for field in attribute:
                if field.startswith("delegate."):
                    _, field_name = field.split(".")
                    field_changes_info = attribute[field]
                    _, value, _ = field_changes_info
                    data[field_name] = value

        try:
            delegate = Delegate.objects.create(**data)
        except ValueError:
            return None

        return delegate


class PrivConfAttrsChangeHelper(ChangeHelper):
    u"""
    Обработчик изменений доп.аттрибутов льгот PrivilegeConfirmationAttributes.
    При проверки изменений атрибутов обладателя льготы, вжно что пришло в
    теге Type, если пришел PrivilegeOwnerEnum.DELEGATE,
    то теги данных будут заполнены и их нужно проверять на изменение,
    иаче проверяем только на изменение тип представителя

    """
    def check_diff(self, conc_object, save_object):
        if conc_object.privilege_owner != PrivilegeOwnerEnum.DELEGATE:
            instanse_model = save_object or conc_object
            field, type_value, _ = ('privilege_owner', int, 'Type')
            conc_value = recursive_getattr(conc_object,
                                           field.replace('.', '__'))
            save_value = recursive_getattr(save_object,
                                           field.replace('.', '__'))
            compare_func = map_changes.get_function_compare(type_value)
            if not compare_func(conc_value, save_value):
                self._handling_change(field, type_value, conc_value,
                                      save_value, instanse_model)
        else:
            super(PrivConfAttrsChangeHelper, self).check_diff(conc_object, save_object)


class LipetskUpdateProxy(UpdateProxy):
    """
    Сравниваем только с льготами с портала
    """

    def _create_delegate(self):
        return LipetskDelegateProxy(self._delegate_data).get()

    def _create_children(self):
        return LipetskChildrenProxy(
            self._child_data, self._delegate_data,
            self._adaptation_program_type).get()

    def _check_benefits(self):
        conc_benefits = self._create_benefits()
        declarationprivilege_set = get_decl_priv(self.declaration, True)
        benefits = map(lambda x: x.privilege.id, declarationprivilege_set)

        # 1. Основная модель привелегий
        self.chek_model('DeclarationPrivilege', conc_benefits, benefits)

        # 2. Модель доп.аттрибут о льготах (липецкая, внутри плагина)

        new_priv_conf_attrs = self._create_privilegeconfattributes()
        priv_conf_attrs = get_priv_data(self.declaration)
        if new_priv_conf_attrs:
            self.chek_model(
                'PrivilegeConfirmationAttributes',
                new_priv_conf_attrs, priv_conf_attrs)

    def _create_privilegeconfattributes(self):
        u"""Получение прокси класса по доп.аттрибутам льгот."""
        declaration_privilege_set = get_decl_priv(self.declaration)
        if declaration_privilege_set:
            return PrivConfirmAttrsProxy(
                self._child_data,
                self.declaration,
                declaration_privilege_set[0]
            ).get()
        else:
            privilege_id = self._child_data.Benefits.Benefit[0].name
            return PrivConfirmAttrsProxy(
                self._child_data,
                self.declaration,
                DeclarationPrivilege(
                    declaration=self.declaration,
                    privilege_id=privilege_id)
            ).get()


class LipetskUpdateApplicationProxy(UpdateApplicationProxy):
    """
    Перекрываем тк нужно применять изменения только
    для "портальных" льгот
    """
    DEFAULT_BEHAVIOR_CLS = LipetskUpdateProxy
